from functools import partial
import json
from random import randint, choice
import string
import unittest
import httpretty
import mock

from ambisafe import Account, Container, Wallet4Transaction, Client, Crypt


def cosign_by_user(self, password, transaction, account_id, currency='BTC'):
    account = self.get_account(account_id, currency)
    container = account.user_container
    private_key = container.get_private_key(password)
    sign = partial(container.sign, private_key=private_key)
    transaction.user_signatures = map(sign, transaction.sighashes)
    return transaction

Client.cosign_by_user = cosign_by_user


class ClientTestCase(unittest.TestCase):
    id = 'test_{}'.format(randint(1, 100000000))
    _get_account_response = '{"account":{"externalId":"%s","currencyFamilySymbol":"BitcoinFamily","securitySchemaName' \
                            '":"Wallet4","address":"3MLbmeUjhiU8Artq2Xp3cdKZzXmocQahpy"},"containers":{"USER":{"publi' \
                            'cKey": "034a94cacac4327feb793047c514b256b326c3c474d73c861407a8709f9901039e", "data": "a0' \
                            'b0cbf2c2697f5141041da8a012149dc4cd82df6f43be8cfc58342ba8e663722178509b667217f2c990ec24ff' \
                            'aeb2ed", "salt": "ca20faef-ac3f-40a6-99c0-500855c03207", "iv": "ff55e03b11dc43adf839c3ae' \
                            'e3632b36"},"OPERATOR":{"publicKey": "034a94cacac4327feb793047c514b256b326c3c474d73c86140' \
                            '7a8709f9901039e", "data": "a0b0cbf2c2697f5141041da8a012149dc4cd82df6f43be8cfc58342ba8e66' \
                            '3722178509b667217f2c990ec24ffaeb2ed", "salt": "ca20faef-ac3f-40a6-99c0-500855c03207", "i' \
                            'v": "ff55e03b11dc43adf839c3aee3632b36"}}}' % id
    _build_transaction_response = '{"hex":"0100000001d2c7543329cc67caf16bdbe2fec1de29f2119c902cb1fc271ab0312e8f4c842b' \
                                  '0000000091000000004c8b5321023acc402315f02158304d2ff6fdb3e08edbc1e052100595c4873b08' \
                                  '73fd36a33a2102a98e0db19265f8da57797384a2cbbb8db33cdf147a6780349a67506d64de2dfb2102' \
                                  '787591a1671fc234f4bd809ed65ac6ab27b63713084e81a295bc9088515e13cc2103c65efb073846e1' \
                                  'c1e3a56f08c29ce79d7c9f5dab5e2c09490985a75c6779cd1c54aeffffffff04ac0200000000000047' \
                                  '5121030c28fe5f18c699a30b62b7d67de92b7493658f2bab2a1806fbf5f0ea7f28e8db21023c3409dd' \
                                  'c09896cdead6d5949b093cd8c94bd18c1fcc5370415ee276f399676852ae22020000000000001976a9' \
                                  '14946cb2e08075bcbaf157e47bcb67eb2b2339d24288ac220200000000000017a9145261efc042151b' \
                                  '5e595797605862128a9842a03487980c00000000000017a914d78608c18942160a642b05733b7786e2' \
                                  'a8d843398700000000","fee":"0.0001 BTC","sighashes":["b3cfce1850fcd9fafa6be899398bb' \
                                  '82327944fbe192b7f43ba1603d6141abb3e"]}'
    _transaction_ok = '{"state":"sending","transaction":"0100000001d2c7543329cc67caf16bdbe2fec1de29f2119c902cb1fc271a' \
                      'b0312e8f4c842b00000000fd6a0100483045022100cb22b4a40b5c7520c1ed808f9514ff29a47b396c8de527758a24' \
                      '8e94fdedd546022001c90fd9fec4ad6188dec59370e47e5fbc90b8623d1effd9d9e1aa83db1f5d3501493046022100' \
                      'e83e4feaa680f5f476116abf6027eb63f005d6bdba3bf17ac0d73bbf5a209407022100bd9d525b7dfe2bc20edcfed0' \
                      'f0e6252f9de8ce0d760cb9fbcb8c153ea21debfe01483045022100a944c93adc0d7a0adf021174bb3fd40921a0b023' \
                      '3ed4c51286eac886346122a2022073c0627906e372d80c046dc8f283c09ddf9aa167ccbdf56019269de9b123713401' \
                      '4c8b5321023acc402315f02158304d2ff6fdb3e08edbc1e052100595c4873b0873fd36a33a2102a98e0db19265f8da' \
                      '57797384a2cbbb8db33cdf147a6780349a67506d64de2dfb2102787591a1671fc234f4bd809ed65ac6ab27b6371308' \
                      '4e81a295bc9088515e13cc2103c65efb073846e1c1e3a56f08c29ce79d7c9f5dab5e2c09490985a75c6779cd1c54ae' \
                      'ffffffff04ac02000000000000475121030c28fe5f18c699a30b62b7d67de92b7493658f2bab2a1806fbf5f0ea7f28' \
                      'e8db21023c3409ddc09896cdead6d5949b093cd8c94bd18c1fcc5370415ee276f399676852ae220200000000000019' \
                      '76a914946cb2e08075bcbaf157e47bcb67eb2b2339d24288ac220200000000000017a9145261efc042151b5e595797' \
                      '605862128a9842a03487980c00000000000017a914d78608c18942160a642b05733b7786e2a8d843398700000000",' \
                      '"transactionHash":"b9beff30a956f37180d7b2e7245677d8b2d9f7728433c7693ee33ac48bd8b0f3"}'

    @classmethod
    def setUpClass(cls):
        cls.client = Client('http://localhost:8080', 'test', 'demo', 'demo')

    @httpretty.activate
    def test_create_account_wallet4(self):
        httpretty.register_uri(httpretty.POST, 'http://localhost:8080/accounts',
                               body=self._get_account_response)
        user_container = Container.generate('test')
        operator_container = Container.generate(self.client.secret)
        result = self.client.create_wallet4_account(self.id,
                                                    user_container=user_container,
                                                    operator_container=operator_container,
                                                    currency='BTC')
        self.assertIsInstance(result, Account)

    @httpretty.activate
    def test_create_account_simple(self):
        httpretty.register_uri(httpretty.POST, 'http://localhost:8080/accounts',
                               body=self._get_account_response)
        result = self.client.create_simple_account(self.id, currency='BTC')
        self.assertIsInstance(result, Account)

    @httpretty.activate
    def test_update_account_wallet4(self):
        httpretty.register_uri(httpretty.PUT, 'http://localhost:8080/accounts',
                               body=self._get_account_response)
        user_container = Container.generate('test')
        result = self.client.update_wallet4_account(self.id, user_container=user_container,
                                                    operator_container=Container.generate(self.client.secret),
                                                    currency='BTC')
        self.assertIsInstance(result, Account)

    @httpretty.activate
    def test_get_balance(self):
        url = 'http://localhost:8080/balances/BTC/{external_id}'.format(external_id=self.id)
        httpretty.register_uri(httpretty.GET, url,
                               body=json.dumps({u'address': u'38gP2E6Fj2s3sx63urYPpSrgTwupHGvKsK',
                                                u'balance': u'0.0011086',
                                                u'balanceInSatoshis': u'110860',
                                                u'currencySymbol': u'BTC'}))
        balance = self.client.get_balance(self.id, 'BTC')
        self.assertEqual(0.0011086, balance)

    @httpretty.activate
    def test_get_account(self):
        url = 'http://localhost:8080/accounts/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.GET, url, body=self._get_account_response)
        account = self.client.get_account(self.id, 'BTC')
        self.assertIsInstance(account, Account)

    @httpretty.activate
    def test_build_transaction(self):
        url = 'http://localhost:8080/transactions/build/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.POST, url, body=self._build_transaction_response)
        transaction = self.client.build_transaction(self.id, 'BTC', '39Ccf2Hr2ns58ugt1mRZw7tKA3AhZu41EJ', 0.00001)
        self.assertIsInstance(transaction, Wallet4Transaction)

    @httpretty.activate
    def test_cosign(self):
        url = 'http://localhost:8080/transactions/build/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.POST, url, body=self._build_transaction_response)
        url = 'http://localhost:8080/accounts/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.GET, url, body=self._get_account_response)

        transaction = self.client.build_transaction(self.id, 'BTC', '39Ccf2Hr2ns58ugt1mRZw7tKA3AhZu41EJ', 0.00001)
        transaction = self.client.sign_wallet4_transaction(transaction, self.id, 'BTC')
        self.assertIsInstance(transaction, Wallet4Transaction)

    @httpretty.activate
    def test_cosign_and_submit(self):
        url = 'http://localhost:8080/transactions/build/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.POST, url, body=self._build_transaction_response)
        url = 'http://localhost:8080/accounts/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.GET, url, body=self._get_account_response)
        url = 'http://localhost:8080/transactions/submit/{external_id}/BTC'.format(external_id=self.id)
        httpretty.register_uri(httpretty.POST, url, body=self._transaction_ok)

        transaction = self.client.build_transaction(self.id, 'BTC', '38EyoyyZC5tyvmhjNyT9YR3ZDoSesCYjPR', 0.00001)
        transaction = self.client.sign_wallet4_transaction(transaction, self.id, 'BTC')
        transaction = self.client.cosign_by_user('test', transaction, self.id, 'BTC')
        response = self.client.submit(self.id, transaction, 'BTC')
        self.assertEqual(response['state'], 'sending')


class CryptTestCase(unittest.TestCase):
    def test_encrypt(self):
        crypt = Crypt('test')
        with mock.patch('ambisafe.crypt.generate_iv') as generate_iv:
            generate_iv.return_value = 'ff55e03b11dc43adf839c3aee3632b36'.decode('hex')
            iv, data = crypt.encrypt('8a3167b6032285a9fd89fcf9110d51ce1cffaf0eb21bc316560d0e510ebac7cd',
                                     'ca20faef-ac3f-40a6-99c0-500855c03207')
        self.assertEqual(data, 'a0b0cbf2c2697f5141041da8a012149dc4cd82df6f43be8c'
                               'fc58342ba8e663722178509b667217f2c990ec24ffaeb2ed')

    def test_decrypt(self):
        crypt = Crypt('test')
        private_key = crypt.decrypt('a0b0cbf2c2697f5141041da8a012149dc4cd82df6f43be8c'
                                    'fc58342ba8e663722178509b667217f2c990ec24ffaeb2ed',
                                    'ca20faef-ac3f-40a6-99c0-500855c03207',
                                    'ff55e03b11dc43adf839c3aee3632b36')
        self.assertEqual(private_key, '8a3167b6032285a9fd89fcf9110d51ce1cffaf0eb21bc316560d0e510ebac7cd')

if __name__ == '__main__':
    unittest.main()



class AmbisafeError(Exception):
    def __init__(self, message, error):
        self.message = message
        self.error = error

    def __str__(self):
        return self.message


class ClientError(AmbisafeError):
    def __repr__(self):
        return '<AmbisafeClientError error="{}" message="{}">'.format(self.error, self.message)


class ServerError(AmbisafeError):
    def __repr__(self):
        return '<AmbisafeClientError error="{}" message="{}">'.format(self.error, self.message)



from .account import Account, Wallet4Account, SimpleAccount, CurrencyIssuerAccount
from .client import Client
from .exc import AmbisafeError, ServerError, ClientError
from .transactions import Wallet4Transaction, RecoveryTransaction, GrantTransaction
from .containers import Container
from .crypt import Crypt, KEY_LENGTH

__all__ = ['Client', 'ServerError', 'AmbisafeError', 'ClientError',
           'Account', 'Wallet4Account', 'SimpleAccount', 'CurrencyIssuerAccount', 'Container',
           'Wallet4Transaction', 'RecoveryTransaction',
           'GrantTransaction', 'Crypt', 'KEY_LENGTH']



import logging
from uuid import uuid4
from ecdsa import SECP256k1, SigningKey, util
from pybitcointools import compress, privkey_to_pubkey
from ambisafe.crypt import Crypt

logger = logging.getLogger('ambisafe')


class Container(object):
    def __init__(self, public_key, data, iv, salt):
        """
        Creates container.
        :param public_key: Public key
        :param data: Encrypted private_key
        :param iv: IV
        :param salt: Salt
        :return: Container
        """
        self.publicKey = public_key
        self.data = data
        self.iv = iv
        self.salt = salt

    @classmethod
    def generate(cls, secret):
        """
        Generating new container encrypted with secret
        :param secret:
        :type secret: basestring
        :return: Container
        """
        key = SigningKey.generate(curve=SECP256k1)
        private_key = key.to_string().encode('hex')
        compressed_public_key = compress(privkey_to_pubkey(private_key))
        crypt = Crypt(secret)
        salt = str(uuid4())
        iv, encrypted_private_key = crypt.encrypt(private_key, salt)
        return cls(compressed_public_key, encrypted_private_key, iv, salt)

    @classmethod
    def from_server_response(cls, publicKey, data, iv, salt):
        """
        Creating Container object from Ambisafe KeyServer response
        :param publicKey:
        :param data:
        :param iv:
        :param salt:
        :return:
        """
        return cls(publicKey, data, iv, salt)

    def get_private_key(self, secret):
        """
        Get decrypted private key from Container using secret
        :param secret:
        :return: str
        """
        crypt = Crypt(secret)
        private_key = crypt.decrypt(self.data, self.salt, self.iv)
        if not private_key:
            raise ValueError('Wrong secret')
        return private_key

    def sign(self, message, private_key):
        """
        Sign message with private key
        :param message:
        :param private_key:
        :return: str
        """
        key = SigningKey.from_string(private_key.decode('hex'), curve=SECP256k1)
        sig = key.sign_digest(message.decode('hex'), sigencode=util.sigencode_der) + '01'.decode('hex')
        logger.debug('pub: {} priv: {} sighash: {} signed: {}'
                     .format(self.publicKey, private_key, message, sig.encode('hex')))
        return sig.encode('hex')

    def __getitem__(self, item):
        return self.__dict__[item]

    def as_response(self):
        """
        Get container dict in response format
        :return: dict
        """
        return self.__dict__

    def as_request(self):
        """
        Get container dict in request format
        :return: dict
        """
        container = self.__dict__
        # work around different request and response formats
        container['public_key'] = container.pop('publicKey')
        return container



from base64 import b64encode
import calendar
from datetime import datetime
import logging
import json
from hashlib import sha512
import hmac
import re
import requests

from .account import Account
from .containers import Container
from .exc import ClientError, ServerError
from .transactions import BaseTransaction, Wallet4Transaction, RecoveryTransaction, GrantTransaction

logger = logging.getLogger('ambisafe')


class Client(object):
    def __init__(self, ambisafe_server, secret, api_key, api_secret):
        if not re.match(r'^(http|https)://', ambisafe_server):
            raise ValueError('ambisafe_server should be http/https URI')
        self.ambisafe_server = ambisafe_server
        self.secret = secret
        self.api_key = api_key
        self.api_secret = api_secret

    def create_simple_account(self, account_id, currency='BTC'):
        """
        Create account with Simple security schema
        :param account_id: int|str
        :param currency: str
        :return:
        """
        logger.debug('Creating account with Simple security schema: account_id: {}, currency: {}'
                     .format(account_id, currency))

        response = self.make_request('POST', '/accounts', json.dumps({
            'id': str(account_id),
            'currency': currency,
            'security_schema': 'Simple',
        }))

        return Account.from_server_response(response)

    def create_wallet4_account(self, account_id, user_container, operator_container, currency='BTC'):
        """
        Create account with Wallet4 security schema
        :param account_id: int|str
        :param user_container: ambisafe.account.Container
        :param operator_container: ambisafe.account.Container
        :param currency: str
        :return:
        """
        logger.debug('Creating account with Wallet4 security schema: account_id: {}, currency: {} '
                     'containers: (user: {}, operator: {})'
                     .format(account_id, currency, user_container, operator_container))

        containers = {
            "USER": user_container.as_request(),
            "OPERATOR": operator_container.as_request(),
        }
        response = self.make_request('POST', '/accounts', json.dumps({
            'id': str(account_id),
            'currency': currency,
            'security_schema': 'Wallet4',
            'containers': containers,
        }))

        return Account.from_server_response(response)

    def create_currency_issuer_account(self, account_id, signatures_required, containers, currency):
        logger.debug('Creating account with CurrencyIssuer security schema: account_id: {}, currency: {} '
                     'containers: {} signatures_required: {}'
                     .format(account_id, currency, containers, signatures_required))
        if any(not isinstance(container, Container) for container in containers):
            raise ValueError("containers should be a list of Container instances")

        containers = [dict(role=str(i), **container.__dict__) for i, container in enumerate(containers)]
        response = self.make_request('POST', '/accounts', json.dumps({
            'id': str(account_id),
            'currency': currency,
            'security_schema': 'CurrencyIssuer',
            'signatures_required': signatures_required,
            'containers': containers,
        }))
        return Account.from_server_response(response)

    def update_wallet4_account(self, account_id, user_container, operator_container,
                               currency='BTC', regenerate_server_container=False):
        containers = {
            "USER": user_container.as_request(),
            "OPERATOR": operator_container.as_request(),
        }
        return Account.from_server_response(self.make_request('PUT', '/accounts', json.dumps({
            'id': account_id,
            'currency': currency,
            'security_schema': 'Wallet4',
            'containers': containers,
            'regenerate_server_container': regenerate_server_container,
        })))

    def get_balance(self, account_id, currency='BTC'):
        return float(
            self.make_request('GET', '/balances/{currency}/{external_id}'
                              .format(currency=currency, external_id=account_id))['balance']
        )

    def get_account(self, account_id, currency='BTC'):
        return Account.from_server_response(
            self.make_request('GET', '/accounts/{external_id}/{currency}'
                              .format(external_id=account_id, currency=currency))
        )

    def build_transaction(self, account_id, currency, destination, amount):
        body = {
            "destination": destination,
            "amount": amount,
        }
        return Wallet4Transaction(**self.make_request('POST', '/transactions/build/{external_id}/{currency}'
                                                              .format(external_id=account_id, currency=currency),
                                                      body=json.dumps(body)))

    def submit(self, account_id, transaction, currency):
        if not isinstance(transaction, BaseTransaction):
            raise ValueError('transaction should be instance of BaseTransaction class')
        return self.make_request('POST', '/transactions/submit/{external_id}/{currency}'
                                 .format(external_id=account_id, currency=currency),
                                 body=transaction.to_json())

    def sign_wallet4_transaction(self, transaction, account_id, currency):
        account = self.get_account(account_id, currency)
        return account.sign(transaction, 'OPERATOR', self.secret)

    def cosign_wallet4_and_submit(self, transaction, account_id, currency='BTC'):
        transaction = self.sign_wallet4_transaction(transaction, account_id, currency)
        return self.submit(account_id, transaction, currency)

    def build_recovery_transaction(self, account_id, currency, old_address):
        response = self.make_request(
            'POST',
            '/transactions/build_recovery/{external_id}/{currency}/{address}'
            .format(external_id=account_id, currency=currency, address=old_address)
        )
        logger.debug('response: {}'.format(response))
        return RecoveryTransaction(response['operator'], response['recovery_transaction'], response['account_id'])

    def cosign_and_recovery(self, transaction, account_id, currency='BTC'):
        transaction = self.sign_wallet4_transaction(transaction, account_id, currency)
        return self.submit(account_id, transaction, currency)

    def build_grant_transaction(self, account_id, currency, destination, amount, signatures_required):
        body = {
            "destination": destination,
            "amount": str(amount),
            "signatures_required": str(signatures_required),
            "operation_type": "grant"
        }
        transaction = self.make_request('POST',
                                        '/transactions/build/{id}/{currency}'.format(id=account_id, currency=currency),
                                        json.dumps(body))
        return GrantTransaction(**transaction)

    def make_request(self, method, uri, body=None):
        url = self.ambisafe_server + uri
        utctime = datetime.utcnow()
        nonce = int(calendar.timegm(utctime.timetuple()) * 10 ** 3
                    + utctime.microsecond * 10 ** -3)
        message = "{}\n{}\n{}\n".format(nonce, method, url)
        if body:
            message += body
        digest = hmac.new(self.api_secret, msg=message, digestmod=sha512)
        signature = b64encode(digest.digest()).replace('\n', '')
        headers = {
            'API-key': self.api_key,
            'signature': signature,
            'timestamp': nonce,
            'Accept': 'application/json'
        }
        if method in ['POST', 'PUT']:
            headers['Content-Type'] = 'application/json'
        logger.debug('Request to ambisafe KeyServer: method: {}, url: "{}", headers: {}, data: {}'
                     .format(method, url, headers, body))
        response = requests.request(method, url, headers=headers, data=body)
        logger.debug('Response from ambisafe KeyServer: status: {}, text: {}'
                     .format(response.status_code, response.text))
        try:
            response_data = response.json()
        except ValueError, e:
            # ValueError is parent of JSONDecodeError
            raise ServerError(e.message, '')

        if not response.ok:
            if 400 <= response.status_code < 500:
                raise ClientError(response_data['message'], response_data['error'])
            elif 500 <= response.status_code < 600:
                raise ServerError(response_data['message'], response_data['error'])

        return response_data



import os
from setuptools import setup, find_packages

VERSION = '0.7'

BASEDIR = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(BASEDIR, 'README.rst')).read()

setup(
    name='ambisafe',
    version=VERSION,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "requests >= 2.7.0",
        "pyOpenSSL >= 0.15.1",
        "ecdsa >= 0.13",
        "pycrypto >= 2.6.1",
        "pybitcointools == 1.1.15"
    ],
    url='https://bitbucket.org/ambisafe/client-python',
    download_url='https://bitbucket.org/ambisafe/client-python/get/{0}.zip'
        .format(VERSION),
    author='Anton Simernia',
    author_email='anton.simernia@ambisafe.co',
    keywords=['ambisafe', 'bitcoin'],
    description='Ambisafe server client library',
    long_description=README,
    classifiers=[
        'Intended Audience :: Developers',
    ],
    test_suite='test.test',
    setup_requires=[
        "flake8",
        "nose>=1.0",
        "coverage",
        "mock"
    ]
)
